#include "stdafx.h"
#include "testmemmap.h"


constexpr const int MAX_EDIT_TEXT{100};
constexpr const TCHAR* g_MMFName{_T("MMFTEST")};
constexpr const TCHAR* g_MMFMutexName{_T("MMFTEST_MUTEX")};

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


#pragma warning(suppress: 26440 26433)
BEGIN_MESSAGE_MAP(CTestmemmapApp, CWinApp) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
END_MESSAGE_MAP()


#pragma warning(suppress: 26426)
CTestmemmapApp theApp;


BOOL CTestmemmapApp::InitInstance()
{
  //seed the random number generator
#pragma warning(suppress: 26472)
  srand(static_cast<unsigned int>(time(nullptr)));

  //Form a path to the test txt file we will be operating on
  CString sTestFile;
  const DWORD dwGMF{GetModuleFileName(nullptr, sTestFile.GetBuffer(_MAX_PATH), _MAX_PATH)};
  sTestFile.ReleaseBuffer();
  if (dwGMF == 0)
  {
    CString sMsg;
    sMsg.Format(_T("Failed in call to GetModuleFileName. Error:%u"), GetLastError());
    AfxMessageBox(sMsg);
    return FALSE;
  }
  CString sDrive;
  CString sDir;
  _tsplitpath_s(sTestFile, sDrive.GetBuffer(_MAX_DRIVE), _MAX_DRIVE, sDir.GetBuffer(_MAX_DIR), _MAX_DIR, nullptr, 0, nullptr, 0);
  sDir.ReleaseBuffer();
  sDrive.ReleaseBuffer();
  _tmakepath_s(sTestFile.GetBuffer(_MAX_PATH), _MAX_PATH, sDrive, sDir, _T("input"), _T("txt"));
  CString sUpperCaseTestFile;
  _tmakepath_s(sUpperCaseTestFile.GetBuffer(_MAX_PATH), _MAX_PATH, sDrive, sDir, _T("outputUpperCase"), _T("txt"));
  sUpperCaseTestFile.ReleaseBuffer();
  CString sLowerCaseTestFile;
  _tmakepath_s(sLowerCaseTestFile.GetBuffer(_MAX_PATH), _MAX_PATH, sDrive, sDir, _T("outputLowerCase"), _T("txt"));
  sLowerCaseTestFile.ReleaseBuffer();

  //First test out the class when we want to map a file
  //You might want to step through this code to test it out
  CMemMapFile mmf;
  BOOL bSuccess{mmf.MapFile(sTestFile, FALSE)};
  VERIFY(bSuccess);
  LPVOID pData{mmf.Open()}; //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(pData != nullptr);
  bSuccess = mmf.Flush();
  VERIFY(bSuccess);
#pragma warning(suppress: 26160)
  bSuccess = mmf.Close();
  VERIFY(bSuccess);
  mmf.UnMap();

  //Now a real demonstration of Memory mapped file usage. Make the sample input.txt file 
  //uppercase. Error checking is ignored / implemented by using VERIFY for demonstration purposes
  //The real power of memory mapped files is that this demonstration would work just as efficiently 
  //with a 100 MB file as it does with just a small text file. Note that because these samples uses 
  //CharUpperBuffA and ChatLowerBuffA, the file size must be less than 4GB since we cast to a DWORD 
  //sizes parameter 
  bSuccess = CopyFile(sTestFile, sUpperCaseTestFile, FALSE);
  VERIFY(bSuccess);
  CMemMapFile mmf3;
  bSuccess = mmf3.MapFile(sUpperCaseTestFile);
  VERIFY(bSuccess);
  pData = mmf3.Open();
#pragma warning(suppress: 26477)
  ATLASSUME(pData != nullptr);
  CFileStatus fs;
  bSuccess = CFile::GetStatus(sTestFile, fs);
  VERIFY(bSuccess);
  VERIFY(fs.m_size <= ULONG_MAX);
#pragma warning(suppress: 26472)
  CharUpperBuffA(static_cast<LPSTR>(pData), static_cast<DWORD>(fs.m_size));
#pragma warning(suppress: 26160)
  bSuccess = mmf3.Close();
  VERIFY(bSuccess);
  mmf3.UnMap();
  CString sMsg;
  sMsg.Format(_T("The file %s should be the uppercase version of %s"), sUpperCaseTestFile.GetString(), sTestFile.GetString());
  AfxMessageBox(sMsg);

  //Also try lowercase version
  bSuccess = CopyFile(sTestFile, sLowerCaseTestFile, FALSE);
  VERIFY(bSuccess);
  CMemMapFile mmf4;
  bSuccess = mmf4.MapFile(sLowerCaseTestFile);
  VERIFY(bSuccess);
  pData = mmf4.Open();
#pragma warning(suppress: 26477)
  ATLASSUME(pData != nullptr);
  bSuccess = CFile::GetStatus(sTestFile, fs);
  VERIFY(bSuccess);
  VERIFY(fs.m_size <= ULONG_MAX);
#pragma warning(suppress: 26472)
  CharLowerBuffA(static_cast<LPSTR>(pData), static_cast<DWORD>(fs.m_size));
#pragma warning(suppress: 26160)
  bSuccess = mmf4.Close();
  VERIFY(bSuccess);
  mmf4.UnMap();
  sMsg.Format(_T("The file %s should be the lowercase version of %s"), sLowerCaseTestFile.GetString(), sTestFile.GetString());
  AfxMessageBox(sMsg);

  //Finally a demonstration of the shared memory capabilities of 
  //memory mapped files. For this demo, we will bring up a UI which
  //will take whatever is put into its edit box and transfer it into 
  //shared memory. Then if you bring up another version of testmemmap, 
  //you will see this value being reflecting into it.
  //periodically
  CDialog1 dlg;
  dlg.DoModal();

#pragma warning(suppress: 26167)
  return FALSE;
}


#pragma warning(suppress: 26455)
CDialog1::CDialog1(CWnd* pParent) : CDialog{CDialog1::IDD, pParent},
                                    m_nTimer{0},
                                    m_pThread1{nullptr},
                                    m_pThread2{nullptr},
                                    m_bWantStop{false}
{
}

void CDialog1::DoDataExchange(CDataExchange* pDX)
{
  //Let the base class do its thing
  __super::DoDataExchange(pDX);

  DDX_Control(pDX, IDC_EDIT1, m_ctrlEdit);
}

#pragma warning(suppress: 26440 26433)
BEGIN_MESSAGE_MAP(CDialog1, CDialog) //NOLINT(modernize-avoid-c-arrays)
  ON_EN_CHANGE(IDC_EDIT1, &CDialog1::OnChangeEdit1)
  ON_WM_DESTROY()
  ON_WM_TIMER()
END_MESSAGE_MAP()

UINT CDialog1::_MoniterSM(_In_ LPVOID pParam)
{
#pragma warning(suppress: 26429)
  auto pDialog{static_cast<CDialog1*>(pParam)};
  return pDialog->MoniterSM();
}

//A sample worker thread to show how to share a memory mapped file across multiple threads
UINT CDialog1::MoniterSM()
{
  //Open up the existing MMF
  CMemMapFile mmf;
  const BOOL bSuccess{mmf.MapExistingMemory(g_MMFName, g_MMFMutexName, (MAX_EDIT_TEXT + 1) * sizeof(TCHAR))};
  VERIFY(bSuccess);

  while (!m_bWantStop)
  {
    //Just modify the data with random text
    CString sText;
    sText.Format(_T("Modified from Thread %u, Random data:%d"), GetCurrentThreadId(), rand());
    LPVOID lpData{mmf.Open()};
    if (lpData != nullptr)
    {
      TRACE(_T("Thread %u update\n"), GetCurrentThreadId());
      _tcscpy_s(static_cast<TCHAR*>(lpData), MAX_EDIT_TEXT + 1, sText);
#pragma warning(suppress: 26110)
      mmf.Close();
    }
    else
    {
      ASSERT(FALSE);
    }
    Sleep(100);
  }
#pragma warning(suppress: 26167)
  return 0;
}

BOOL CDialog1::OnInitDialog()
{
  //Let the base class do its thing
  __super::OnInitDialog();

  //Limit the text to the size that can be held in the MMF
  m_ctrlEdit.LimitText(MAX_EDIT_TEXT);

  //Check to see if we are the first instance
  //of this program running. In the process testing
  //the MapExistingMemory method
  if (m_mmf.MapExistingMemory(g_MMFName, g_MMFMutexName, (MAX_EDIT_TEXT+1)*sizeof(TCHAR)))
    AfxMessageBox(_T("This program is already running"));
  else
  {
    AfxMessageBox(_T("This program is the first instance"));

    //map the shared memory
    if (m_mmf.MapMemory(g_MMFName, g_MMFMutexName, (MAX_EDIT_TEXT+1)*sizeof(TCHAR)))
    {
      //Set the initial text which will go into the MMF
      m_ctrlEdit.SetWindowText(_T("Initial Text"));
    }
    else
    {
      AfxMessageBox(_T("Failed to map shared memory"));
      PostMessage(WM_CLOSE);
    }
  }

  //Create a 1 second timer to allow us to update the edit box
  m_nTimer = SetTimer(1, 1000, nullptr);

#ifdef _DEBUG
  ASSERT(m_pThread1 == nullptr);
  m_pThread1 = AfxBeginThread(_MoniterSM, this, THREAD_PRIORITY_NORMAL, CREATE_SUSPENDED);
  ASSERT(m_pThread1 != nullptr);
  m_pThread1->m_bAutoDelete = FALSE;
  m_pThread1->ResumeThread();

  ASSERT(m_pThread2 == nullptr);
  m_pThread2 = AfxBeginThread(_MoniterSM, this, THREAD_PRIORITY_NORMAL, CREATE_SUSPENDED);
  ASSERT(m_pThread2 != nullptr);
  m_pThread2->m_bAutoDelete = FALSE;
  m_pThread2->ResumeThread();
#endif

  return TRUE;
}

void CDialog1::OnChangeEdit1()
{
  CString sText;
  m_ctrlEdit.GetWindowText(sText.GetBuffer(MAX_EDIT_TEXT), MAX_EDIT_TEXT);
  sText.ReleaseBuffer();
  LPVOID lpData{m_mmf.Open()};
  if (lpData != nullptr)
  {
    TRACE(_T("CDialog1::OnChangeEdit1 update\n"));
    _tcscpy_s(static_cast<TCHAR*>(lpData), MAX_EDIT_TEXT + 1, sText);
#pragma warning(suppress: 26110 26167)
    m_mmf.Close();
  }
}

#pragma warning(suppress: 26434)
void CDialog1::OnDestroy()
{
  KillTimer(m_nTimer);
  m_nTimer = 0;

  //Wait for the worker threads to exit
  if (m_pThread1 != nullptr)
  {
    m_bWantStop = true;
    WaitForSingleObject(m_pThread1->m_hThread, INFINITE);
    delete m_pThread1;
    m_pThread1 = nullptr;
  }
  if (m_pThread2 != nullptr)
  {
    m_bWantStop = true;
    WaitForSingleObject(m_pThread2->m_hThread, INFINITE);
    delete m_pThread2;
    m_pThread2 = nullptr;
  }


  //Let the base class do its thing
  __super::OnDestroy();
}

#pragma warning(suppress: 26434)
void CDialog1::OnTimer(UINT_PTR /*nIDEvent*/)
{
  //Retreive the contents of the MMF
  CString sNewText;
  LPVOID lpData{m_mmf.Open()};
  if (lpData != nullptr)
  {
    TRACE(_T("CDialog1::OnTimer update\n"));
    sNewText = static_cast<TCHAR*>(lpData);
#pragma warning(suppress: 26110)
    m_mmf.Close();
  }

  //update the edit control if the text has changed
  CString sCurrentText;
  m_ctrlEdit.GetWindowText(sCurrentText);
#pragma warning(suppress: 26167)
  if (sCurrentText != sNewText)
    m_ctrlEdit.SetWindowText(sNewText);
}
